import mock
import re
import unittest
import responses
from django.contrib.auth.models import User
from django.test.client import RequestFactory
from dbmi_client import authz
from dbmi_client.settings import dbmi_settings


class TestAuthz(unittest.TestCase):

    # User JWT
    fake_jwt = 'somefakejwtalsdijlaiwjdlijawlidjasdhgashgd'

    @classmethod
    def setUpClass(cls):

        # Patch JWT checker
        cls.jwt_patcher = mock.patch('dbmi_client.authn.validate_rs256_jwt')
        cls.mock_jwt = cls.jwt_patcher.start()
        cls.mock_jwt.return_value = True

        # Set a user and request factory
        cls.rf = RequestFactory()
        cls.user = User.objects.create_user(username='testuser', email='testuser@dbmiauth.local', password='top_secret')

        # Set the authorization server URL
        cls.authz_url_pattern = re.compile(dbmi_settings.AUTHZ_URL + r'.*')

    @classmethod
    def tearDownClass(cls):

        # Disable patcher
        cls.jwt_patcher.stop()

    def build_request(self, path, method='get'):

        # Build the request
        request = getattr(self.rf, method)(path, HTTP_AUTHORIZATION=TestAuthz.fake_jwt)

        # Add the user
        request.user = self.user
        request.COOKIES[dbmi_settings.JWT_COOKIE_NAME] = 'JWT {}'.format(TestAuthz.fake_jwt)

        return request

    def setUp(self):

        # Setup each test
        pass

    @responses.activate
    @mock.patch('dbmi_client.authz.has_a_permission')
    def test_authz_mock_(self, mock_has_a_permission):

        # Set the response
        mock_has_a_permission.return_value = True

        # Build the request
        request = self.build_request('/some/page/')

        # Run it
        has_authz = authz.has_a_permission(request, self.user.email, 'item', ['admin', 'read'], check_parents=True)

        # Check it
        self.assertTrue(has_authz)

    @responses.activate
    def test_authz_no_response(self):

        # Build the request
        request = self.build_request('/some/page/')

        # Build the response handler
        responses.add(responses.GET,
                      self.authz_url_pattern,
                      json={'error': 'not found'},
                      status=404)

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, 'item', 'admin', check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)

    @responses.activate
    def test_authz_no_perm(self):

        # Build the request
        request = self.build_request('/some/page/')

        # Build the response handler
        responses.add(responses.GET,
                      self.authz_url_pattern,
                      json={'results': [
                          {'item': 'someitem', 'permission': 'admin', 'email': self.user.email}
                      ]},
                      status=200)

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, 'item', 'admin', check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)

    @responses.activate
    def test_authz_wrong_perm(self):

        # Build the request
        request = self.build_request('/some/page/')

        # Build the response handler
        responses.add(responses.GET,
                      self.authz_url_pattern,
                      json={'results': [
                          {'item': 'item', 'permission': 'read', 'email': self.user.email}
                      ]},
                      status=200)

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, 'item', 'admin', check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)

    @responses.activate
    def test_authz_case_insensitive_perm(self):

        # Build the request
        request = self.build_request('/some/page/')

        # Build the response handler
        responses.add(responses.GET,
                      self.authz_url_pattern,
                      json={'results': [
                          {'item': 'item', 'permission': 'admin', 'email': self.user.email}
                      ]},
                      status=200)

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, 'ITEM', 'ADMIN', check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(has_authz)

    @responses.activate
    def test_authz_case_insensitive_perm_1(self):

        # Build the request
        request = self.build_request('/some/page/')

        # Build the response handler
        responses.add(responses.GET,
                      self.authz_url_pattern,
                      json={'results': [
                          {'item': 'ITEM', 'permission': 'ADmIN', 'email': self.user.email}
                      ]},
                      status=200)

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, 'item', 'admin', check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(has_authz)

    @responses.activate
    def test_authz_has_perm(self):

        # Build the request
        request = self.build_request('/some/page/')

        # Build the response handler
        responses.add(responses.GET,
                      self.authz_url_pattern,
                      json={'results': [
                          {'item': 'item', 'permission': 'admin', 'email': self.user.email}
                      ]},
                      status=200)

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, 'item', 'admin', check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(has_authz)

    @responses.activate
    def test_authz_parent_has_perm(self):

        # Build the request
        request = self.build_request('/')

        # Build the response handler
        responses.add(responses.GET,
                      self.authz_url_pattern,
                      json={'results': [
                          {'item': 'parent', 'permission': 'admin', 'email': self.user.email}
                      ]},
                      status=200)

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, 'parent.child', 'admin', check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(has_authz)

    @responses.activate
    def test_authz_grandparent_has_perm(self):

        # Build the request
        request = self.build_request('/')

        # Build the response handler
        responses.add(responses.GET,
                      self.authz_url_pattern,
                      json={'results': [
                          {'item': 'parent', 'permission': 'admin', 'email': self.user.email}
                      ]},
                      status=200)

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, 'parent.child.grandchild', 'admin', check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(has_authz)

    @responses.activate
    def test_authz_parent_wrong_perm(self):

        # Build the request
        request = self.build_request('/')

        # Build the response handler
        responses.add(responses.GET,
                      self.authz_url_pattern,
                      json={'results': [
                          {'item': 'parent', 'permission': 'read', 'email': self.user.email}
                      ]},
                      status=200)

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, 'parent.child', 'admin', check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)

    @responses.activate
    def test_authz_child_perms(self):

        # Build the request
        request = self.build_request('/')

        # Build the response handler
        responses.add(responses.GET,
                      self.authz_url_pattern,
                      json={'results': [
                          {'item': 'parent.child', 'permission': 'admin', 'email': self.user.email}
                      ]},
                      status=200)

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, 'parent', 'admin', check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)

    @responses.activate
    def test_authz_grandchild_perms(self):

        # Build the request
        request = self.build_request('/')

        # Build the response handler
        responses.add(responses.GET,
                      self.authz_url_pattern,
                      json={'results': [
                          {'item': 'parent.child.grandchild', 'permission': 'admin', 'email': self.user.email}
                      ]},
                      status=200)

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, 'parent.child', 'admin', check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)

    @responses.activate
    def test_authz_multiple_no_perms(self):

        # Build the request
        request = self.build_request('/')

        # Build the response handler
        responses.add(responses.GET,
                      self.authz_url_pattern,
                      json={'results': [
                          {'item': 'parent1', 'permission': 'admin', 'email': self.user.email},
                          {'item': 'parent2', 'permission': 'read', 'email': self.user.email}
                      ]},
                      status=200)

        # Build the call
        has_authz = authz.has_a_permission(request=request,
                                           email=self.user.email,
                                           item='parent',
                                           permissions=['admin', 'read'],
                                           check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)

    @responses.activate
    def test_authz_multiple_perms(self):

        # Build the request
        request = self.build_request('/')

        # Build the response handler
        responses.add(responses.GET,
                      self.authz_url_pattern,
                      json={'results': [
                          {'item': 'parent1', 'permission': 'admin', 'email': self.user.email},
                          {'item': 'parent', 'permission': 'read', 'email': self.user.email}
                      ]},
                      status=200)

        # Build the call
        has_authz = authz.has_a_permission(request=request,
                                           email=self.user.email,
                                           item='parent',
                                           permissions=['admin', 'read'],
                                           check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(has_authz)

    @responses.activate
    def test_authz_multiple_perms(self):

        # Build the request
        request = self.build_request('/')

        # Build the response handler
        responses.add(responses.GET,
                      self.authz_url_pattern,
                      json={'results': [
                          {'item': 'parent1', 'permission': 'admin', 'email': self.user.email},
                          {'item': 'parent', 'permission': 'read', 'email': self.user.email}
                      ]},
                      status=200)

        # Build the call
        has_authz = authz.has_a_permission(request=request,
                                           email=self.user.email,
                                           item='parent',
                                           permissions=['admin', 'read'],
                                           check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(has_authz)

    @responses.activate
    def test_authz_grandchild_multiple_perms(self):

        # Build the request
        request = self.build_request('/')

        # Build the response handler
        responses.add(responses.GET,
                      self.authz_url_pattern,
                      json={'results': [
                          {'item': 'parent.child.grandchild', 'permission': 'admin', 'email': self.user.email},
                          {'item': 'parent.child', 'permission': 'read', 'email': self.user.email}
                      ]},
                      status=200)

        # Build the call
        has_authz = authz.has_a_permission(request=request,
                                           email=self.user.email,
                                           item='parent.child',
                                           permissions=['admin', 'read'],
                                           check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(has_authz)

    @responses.activate
    def test_authz_grandchild_multiple_perms_1(self):

        # Build the request
        request = self.build_request('/')

        # Build the response handler
        responses.add(responses.GET,
                      self.authz_url_pattern,
                      json={'results': [
                          {'item': 'parent.child.grandchild', 'permission': 'admin', 'email': self.user.email},
                          {'item': 'parent.child', 'permission': 'read', 'email': self.user.email}
                      ]},
                      status=200)

        # Build the call
        has_authz = authz.has_a_permission(request=request,
                                           email=self.user.email,
                                           item='parent.child',
                                           permissions=['admin', 'write'],
                                           check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)




"""
DBMI Client
"""

__title__ = 'DBMI Client'
__version__ = '0.3.13'
__author__ = 'Bryan Larson'
__license__ = 'BSD 2-Clause'
__copyright__ = 'Copyright 2011-2018 Harvard Medical School Department of Biomedical Informatics'

# Version synonym
VERSION = __version__

# Header encoding (see RFC5987)
HTTP_HEADER_ENCODING = 'iso-8859-1'

# Default datetime input and output formats
ISO_8601 = 'iso-8601'

default_app_config = 'dbmi_client.apps.DBMIClientConfig'


import base64
import json
from enum import Enum

import requests
from furl import furl
from django.conf import settings

from dbmi_client.settings import dbmi_settings
from dbmi_client import authn

import logging
logger = logging.getLogger(dbmi_settings.LOGGER_NAME)


# Set the possible permissions
class FileserviceRole(Enum):
    Readers = '{}__READERS'.format(dbmi_settings.FILESERVICE_GROUP)
    Writers = '{}__WRITERS'.format(dbmi_settings.FILESERVICE_GROUP)
    Downloaders = '{}__DOWNLOADERS'.format(dbmi_settings.FILESERVICE_GROUP)
    Uploaders = '{}__UPLOADERS'.format(dbmi_settings.FILESERVICE_GROUP)
    admins = '{}__ADMINS'.format(dbmi_settings.FILESERVICE_GROUP)

    @classmethod
    def roles(cls):
        return [f.value for f in FileserviceRole]


def create_group():
    """
    Checks Fileservice to ensure it is setup and configured to manage the passed bucket(s)
    with the passed admin(s).
    """
    # Group was not found, create it, specifying passed admins
    data = {
        'name': dbmi_settings.FILESERVICE_GROUP.upper(),
        'users': [{'email': email} for email in dbmi_settings.FILESERVICE_ADMINS],
        'buckets': [{'name': b} for b in dbmi_settings.FILESERVICE_BUCKETS],
    }

    # Make the URL.
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments = ['filemaster', 'groups', '']

    # Make the request
    groups = _request(method='post', url=url.url, data=json.dumps(data))
    if not groups:
        logger.info('Failed to create groups')
        return False

    # Make the request.
    data = {'buckets': [{'name': b} for b in dbmi_settings.FILESERVICE_BUCKETS]}
    for group in groups:

        # Make the URL
        url = furl(dbmi_settings.FILESERVICE_URL)
        url.path.segments = ['filemaster', 'groups', group['id'], '']

        # Make the request
        response = _request(method='put', url=url.url, data=json.dumps(data))
        if response:
            logger.info('Added buckets "{}" to group "{}"'.format(
                dbmi_settings.FILESERVICE_BUCKETS, group['name']
            ))
        else:
            logger.info('Failed to add buckets "{}" to group "{}"'.format(
                dbmi_settings.FILESERVICE_BUCKETS, group['name']
            ))

    return True


def check_group():
    """
    Checks Fileservice to ensure it is setup and configured to manage the files for passed group.
    :return: Whether Fileservice is configured or not
    :rtype: bool
    """
    # Make the URL.
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments = ['filemaster', 'groups', '']

    # Make the request
    groups = _request(method='get', url=url.url)
    if groups is None:
        logger.info('Getting groups failed')
        return False

    # Check for the required group.
    for group in groups:
        if group['name'] in FileserviceRole.roles():
            return True

    return False


def create_archivefile(filename, metadata=None, tags=None):
    """
    Create an ArchiveFile record for the give parameters and returns Fileservice's response.
    :param filename: The filename of the file
    :param metadata: A dictionary of metadata
    :param tags: A list of tags
    :return:
    """
    # Build the request.
    data = {
        'permissions': [
            dbmi_settings.FILESERVICE_GROUP,
        ],
        'filename': filename,
    }

    # Check for and add optional data
    if metadata:
        data['metadata'] = metadata
    if tags:
        data['tags'] = tags

    # Make the URL.
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments = ['filemaster', 'api', 'file', '']

    # Make the request
    file = _request(method='post', url=url.url, data=json.dumps(data))

    return file


def create_archivefile_upload(filename, metadata=None, tags=None, bucket=None, conditions=None):
    """
    Create a file and generate the presigned S3 POST for sending the file to S3. Allows for specification
    of additional parameters for the upload.
    :param filename: The filename of the file
    :param metadata: A dictionary of metadata
    :param tags: A list of tags
    :param bucket: The bucket to which the upload is going
    :param conditions: A list of S3 conditions to pass along when creating the signature for the post
    :return: File UUID, POST data for file upload
    """
    # Check bucket
    if not bucket:
        if dbmi_settings.FILESERVICE_BUCKETS:
            bucket = dbmi_settings.FILESERVICE_BUCKETS[0]
        else:
            raise ValueError('Cannot upload file without bucket specified')

    # Make the request.
    file = create_archivefile(filename=filename, metadata=metadata, tags=tags)

    # Get the UUID.
    uuid = file['uuid']

    # Form the request for the file link
    params = {
        'cloud': 'aws',
        'bucket': bucket,
        'expires': 100
    }

    # Add conditions if passed
    if conditions:
        params['conditions'] = base64.b64encode(json.dumps(conditions).encode()).decode()

    # Make the URL.
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments = ['filemaster', 'api', 'file', uuid, 'post', '']
    url.query.params = params

    # Make the request
    data = _request(method='get', url=url.url)

    return uuid, data


def get_archivefiles(uuids=None):
    """
    Accepts a list of Fileservice UUIDs and returns a list of the ArchiveFile dictionaries
    from Fileservice
    :param uuids: A list of Fileservice UUIDs
    :type uuids: list
    :return: A list of ArchiveFile dictionaries
    :rtype: list
    """
    # Build the request.
    if uuids and type(uuids) is str:
        params = {
            'uuids': uuids
        }
    elif uuids and type(uuids) is list:
        params = {
            'uuids': ','.join(uuids)
        }
    else:
        params = {}

    # Make the URL.
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments = ['filemaster', 'api', 'file', '']
    url.query.params = params

    # Make the request
    files = _request(method='get', url=url.url)
    if files:
        return files
    else:
        return []


def get_archivefile(uuid):
    """
    Accepts a Fileservice UUID and returns an ArchiveFile dictionary from Fileservice
    :param uuid: A Fileservice UUID
    :type uuid: str
    :return: An ArchiveFile dictionaries
    :rtype: dict
    """
    # Make the request.
    return next(iter(get_archivefiles([uuid])), None)


def get_archivefile_url(uuid):
    """
    Retrieves the ArchiveFile Fileservice download URL. Performing a GET to this
    URL will generate a pre-signed URL to download the actual file from the
    storage location (S3 bucket).
    :param uuid: The UUID of the ArchiveFile for which the hash has been requested
    :type uuid: str
    :return: The file's hash
    :rtype: str
    """
    # Make the URL.
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments = ['filemaster', 'api', 'file', uuid, 'download', '']
    return url.url


def get_archivefile_hash(uuid):
    """
    Retrieves the file hash from the storage provider (or Fileservice). This is
    typically an MD5 hash.
    :param uuid: The UUID of the ArchiveFile for which the hash has been requested
    :type uuid: str
    :return: The file's hash
    :rtype: str
    """
    # Make the URL.
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments = ['filemaster', 'api', 'file', uuid, 'filehash', '']

    # Make the request
    hash = _request(method='get', url=url.url)
    return hash


def update_archivefile(uuid, file):
    """
    Accepts details on an ArchiveFile object on Fileservice
    :param uuid: The UUID of the ArchiveFile to update
    :type uuid: str
    :param file: The file dictionary specifying ArchiveFile updates
    :type file: dict
    :return: The updated ArchiveFile dictionary
    :rtype: dict
    """
    # Make the URL.
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments = ['filemaster', 'api', 'file', uuid, '']

    # Make the request
    file = _request(method='patch', url=url.url, data=json.dumps(file))
    return file


def copy_archivefile(uuid, bucket):
    """
    Copy the requested file to the new bucket
    :param uuid: The file's Fileservice UUID
    :param bucket: The destination S3 bucket
    :return: The result of the operation
    """
    # Make the URL.
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments = ['filemaster', 'api', 'file', uuid, 'copy', '']
    url.query.params.add('to', bucket)

    # Make the request
    file = _request(method='post', url=url.url)
    return file


def move_archivefile(uuid, bucket):
    """
    Move the requested file to the new bucket
    :param uuid: The file's Fileservice UUID
    :param bucket: The destination S3 bucket
    :return: The result of the operation
    """
    # Make the URL.
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments = ['filemaster', 'api', 'file', uuid, 'move', '']
    url.query.params.add('to', bucket)

    # Make the request
    file = _request(method='post', url=url.url)
    return file


def delete_archivefile(uuid, location):
    """
    Delete the requested file from the passed location
    :param uuid: The file's Fileservice UUID
    :param location: The file's location to delete (if in multiple locations)
    :return: The result of the operation
    """
    # Make the URL.
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments = ['filemaster', 'api', 'file', uuid, '']
    url.query.params.add('location', location)

    # Make the request
    file = _request(method='delete', url=url.url)
    return file


def uploaded_archivefile(uuid, location_id):
    """
    Informs Fileservice that the ArchiveFile's actual file has been successfully uploaded to the
    storage location (S3 bucket). Fileservice will run routines to inspect the final location of the
    file and ensure its metadata is updated within Fileservice.
    :param uuid: The UUID of the ArchiveFile being uploaded
    :type uuid: str
    :param location_id: The ID of the location generated by the ArchiveFile creation
    :type location_id: str
    :return: The success of the update operation
    :rtype: bool
    """
    # Make the URL.
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments = ['filemaster', 'api', 'file', uuid, 'uploadcomplete', '']
    url.query.params.add('location', location_id)

    # Make the request
    file = _request(method='get', url=url.url)
    return file


def get_archivefile_download_url(uuid):
    """
    Returns a URL by which the file's data can be downloaded from Fileservice's storage location (S3 bucket).
    The URL will point directly to the file's storage location, likely an S3 bucket/key formatted URL.
    Authentication/authorization is embedded into the URL via AWS pre-signed query but the link will have
    limited life before it's validity will expire (typically 24 hours).
    :param uuid: The UUID of the file to download
    :type uuid: str
    :return: The download URL
    :rtype: str
    """
    # Make the URL.
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments = ['filemaster', 'api', 'file', uuid, 'download', '']

    # Make the request
    file = _request(method='get', url=url.url)
    return file['url']


def download_archivefile(uuid):
    """
    Downloads a file from Fileservice's storage location (S3 bucket).
     This downloads the file directly and returns the bytes of said file. This download
    occurs synchronously and is not recommended for files of significant size.
    :param uuid: The UUID of the file to download
    :type uuid: str
    :return: The raw bytes of the file from S3
    :rtype: bytes
    """
    # Get the URL
    url = get_archivefile_download_url(uuid)

    # Request the file from S3 and get its contents.
    response = requests.get(url)

    # Add the content to the FHIR resource as a data element and remove the URL element.
    return response.content


def get_archivefile_proxy_url(uuid):
    """
    Returns a URL by which the file's data can be downloaded from Fileservice's storage location (S3 bucket),
    as proxied through Fileservice. Thus, a requesting entity must be able to properly authenticate/authorize
    with Fileservice.
    :param uuid: The UUID of the file to download
    :type uuid: str
    :return: The proxy download URL
    :rtype: str
    """
    # Make the URL.
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments = ['filemaster', 'api', 'file', uuid, 'proxy', '']
    return url.url


def proxy_archivefile(uuid):
    """
    Downloads a file from Fileservice's storage location (S3 bucket) using the Nginx file proxy
    mechanism. This downloads the file directly and returns the bytes of said file. This download
    occurs synchronously and is not recommended for files of significant size.
    :param uuid: The UUID of the file to download
    :type uuid: str
    :return: The raw bytes of the file from S3
    :rtype: bytes
    """
    # Make the URL.
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments = ['filemaster', 'api', 'file', uuid, 'proxy', '']

    # Make the request
    response = _request(method='get', url=url.url, raw=True)

    # Return the data
    return response.content


def download_archivefiles(uuids):
    """
    Accepts a list of Fileservice UUIDs and builds a download of an archive containing those files.
    The object returned is the raw bytes of the archive containing the requested files. The download
    takes place synchronously and is not recommended for files of significant size.
    :param uuids: A list of UUIDs to include in the archive
    :type uuids: list
    :return: The raw data of the archive
    :rtype: bytes
    """
    # Make the URL.
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments = ['filemaster', 'api', 'file', 'archive', '']
    url.query.params.add('uuids', ','.join(uuids))

    # Make the request
    response = _request(method='get', url=url.url, raw=True)

    # Return the data
    return response.content


def _request(method, url, request=None, raw=False, **kwargs):
    """
    This wraps the logic and error-handling required of making a RESTful request to a service
    :param method: The request method (get, post, put, delete, etc)
    :param url: The URL to make the request to
    :param request: The current request object
    :param raw: Return the raw response object
    :return: object
    """
    # Make the request, catch failures, and return response, if any
    response = None
    try:
        # Make the request
        response = getattr(requests, method)(url, headers=_headers(request), **kwargs)
        response.raise_for_status()

        # Determine what to return
        if raw:
            return response
        else:
            return response.json()

    except Exception as e:
        logger.debug('Request/{} failed -> {}: {}'.format(method, url, getattr(response, 'content', '<empty>')))
        logger.exception('Request error: {}'.format(e), exc_info=True, extra={
            'response': response, 'method': method, 'url': url, **kwargs
        })

    # Determine what to return
    if raw:
        return response
    else:
        return None


def _headers(request=None):
    """
    Returns the headers to use for Fileservice requests. If a Fileservice token is specified in settings,
    this will use that token to sign calls, otherwise, the current user's JWT will be used.
    :param request: The request
    :return: dict
    """
    if dbmi_settings.FILESERVICE_TOKEN:

        # Use the service token from environment
        return {"Authorization": 'Token {}'.format(dbmi_settings.FILESERVICE_TOKEN), 'Content-Type': 'application/json'}

    elif request:

        # Get the JWT
        token = authn.get_jwt(request)

        # Use the service token from environment
        return {"Authorization": 'JWT {}'.format(token), 'Content-Type': 'application/json'}

    else:
        raise ValueError('Cannot properly authenticate service call')


# Indicates the requesting user does not have sufficient permissions for the operation
class FileservicePermissionError(Exception):
    pass


# Indicates the current app's group does not exist
class FileserviceGroupError(Exception):
    pass


# Indicates the current app's targeted bucket does not exist in Fileservice
class FileserviceBucketError(Exception):
    pass



#!/usr/bin/env python
"""
This script is a trick to setup a fake Django environment, since this reusable
app will be developed and tested outside any specifiv Django project.
Via ``settings.configure`` you will be able to set all necessary settings
for your app and run the tests as if you were calling ``./manage.py test``.
"""
import sys
from django_nose import NoseTestSuiteRunner
from django.conf import settings

EXTERNAL_APPS = [
    'django.contrib.admin',
    'django.contrib.admindocs',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.messages',
    'django.contrib.sessions',
    'django.contrib.staticfiles',
    'django.contrib.sitemaps',
    'django.contrib.sites',
]
INTERNAL_APPS = [
    'django_nose',
    'dbmi_client',
]
INSTALLED_APPS = EXTERNAL_APPS + INTERNAL_APPS

if not settings.configured:
    settings.configure(
        DATABASES={
            "default": {
                "ENGINE": "django.db.backends.sqlite3",
                "NAME": ":memory:",
            }
        },
        INSTALLED_APPS=INSTALLED_APPS,
        DBMI_CLIENT_CONFIG={
            "CLIENT": "DBMI-Client",
            "ENVIRONMENT": "prod",
            "AUTH0_TENANTS": "hms-dbmi",
        }
    )


def main(*test_args):
    sys.exit(NoseTestSuiteRunner(verbosity=2, interactive=True).run_tests(test_args))


if __name__ == '__main__':
    main(*sys.argv[1:])



from django.contrib import auth as django_auth

from rest_framework import serializers


class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = django_auth.get_user_model()
        exclude = ['password', ]



from django.contrib import auth as django_auth
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import PermissionDenied
from django.utils.functional import SimpleLazyObject
from django.utils.deprecation import MiddlewareMixin

from dbmi_client.settings import dbmi_settings
from dbmi_client import authn

# Get the app logger
import logging
logger = logging.getLogger(dbmi_settings.LOGGER_NAME)


class DBMIAuthenticationMiddleware(MiddlewareMixin):
    """
    Before loading cached user objects, we want to double-check the JWT to:
    1. Ensure it exists still
    2. Ensure it belongs to the currently cached user
    If any of the above do not pass, do the logout routine

    This middleware is a hybrid of purely JWT authentication and session authentication. The
    initial authentication is performed by inspecting the JWT and then loading/creating the user.
    That user's pk is cached exactly how normal session auth works, and is then consulted from there
    on out to load that user to each request. Once the session expires, the JWT is then looked at
    again to provide authentication. Also, the JWT is still consulted on every request to make sure
    the current user is matched to the current JWT. This check is simply comparing the JWT username/email
    to that of the user instance. If a user session exists and the JWT was swapped for some reason,
    this will detect that change and invalidate the current user session and require another initial
    authentication process.

    When you would use this middleware: This is ideal for instances in which the authentication process
    does some heavy or involved work. If authenticating depends on requests being made to the authorization
    server or resources being pulled from remote sources, this work would be done for every single request,
    and would not be ideal. Doing it upon first auth, and then using the session to determine if the user
    is current or not, minimizes that work, but still ensures JWT defines authentication state.
    """

    def process_request(self, request):
        request.user = SimpleLazyObject(lambda: self.__class__.get_jwt_user(request))

    @staticmethod
    def get_jwt_user(request):

        # Check for a valid token
        token = authn.get_jwt(request)
        if not token or not authn.validate_rs256_jwt(token):
            return AnonymousUser()

        # Get their username
        username = authn.get_jwt_username(request, verify=False)

        # Use the usual routine to get the currently cached user
        user = django_auth.get_user(request)
        if user.is_authenticated:
            logger.debug('Found existing User session: {}'.format(username))

            # A cached user is present. We need to double-check JWT user to ensure
            # they are the same as the cached user.
            username = authn.get_jwt_username(request, verify=False)
            email = authn.get_jwt_email(request, verify=False)
            if username and email:
                if not user.username.lower() == username.lower() or not user.email.lower() == email.lower():
                    logger.debug('User session does not match JWT, logging out')

                    # TODO: Figure out if its necessary to person any session invalidation here
                    return AnonymousUser()

        else:
            logger.debug('No existing User, attempting to login: {}'.format(username))

            # No user is logged in but we have a JWT token. Attempt to authenticate
            # the current JWT and if it succeeds, login and cache the user.
            user = django_auth.authenticate(request, token=token)
            if user and user.is_authenticated:
                logger.debug('User has authenticated: {}'.format(username))

                # Store this user in session
                django_auth.login(request, user)

            else:
                logger.debug('User could not be authenticated: {}'.format(username))
                # Whatever token this user has, it's not valid OR their account would/could not
                # be created, deny permission. This will likely be the case for instances where
                # automatic user creation is disabled and a user with a valid JWT is not being
                # granted an account.
                raise PermissionDenied

        return user


class DBMIUsersAuthenticationMiddleware(DBMIAuthenticationMiddleware):
    """
    Before loading cached user objects, we want to double-check the JWT to:
    1. Ensure it exists still
    2. Ensure it belongs to the currently cached user.
    3. If the cached user is an admin, confirm their permissions by re-running login and authorization sync.
    If any of the above do not pass, do the logout routine

    This middleware is a hybrid of purely JWT authentication and session authentication. The
    initial authentication is performed by inspecting the JWT and then loading/creating the user.
    That user's pk is cached exactly how normal session auth works, and is then consulted from there
    on out to load that user to each request. Once the session expires, the JWT is then looked at
    again to provide authentication. Also, the JWT is still consulted on every request to make sure
    the current user is matched to the current JWT. This check is simply comparing the JWT username/email
    to that of the user instance. If a user session exists and the JWT was swapped for some reason,
    this will detect that change and invalidate the current user session and require another initial
    authentication process.

    What sets this middelware apart from `DBMIAuthenticationMiddleware` is admin users are checked
    with every request to ensure permissions are up-to-date. The additional overhead might be considered
    appropriate to ensure changes to administrator permissions are checked with every request to prevent
    a cached user to have elevated permissions after those permissions might have been revoked at the
    authorization server. Normal users are cached and fetched normally and this do not impose the
    additional overhead of double-checking status.

    When you would use this middleware: This is ideal for instances in which the authentication process
    does some heavy or involved work, but admins should be confirmed on every request, no matter what.
    If authenticating depends on requests being made to the authorization server or resources being pulled
    from remote sources, this work would be done for every single request, and would not be ideal. Doing
    it upon first auth, and then using the session to determine if the user is current or not, minimizes
    that work, but still ensures JWT defines authentication state.
    """

    def process_request(self, request):
        request.user = SimpleLazyObject(lambda: self.__class__.get_jwt_user(request))

    @staticmethod
    def get_jwt_user(request):

        # Use super's implementation
        user = super(DBMIUsersAuthenticationMiddleware).get_jwt_user(request)
        if user:

            # Check if they've been granted admin level privileges
            if user.is_staff or user.is_superuser:

                # Get details
                token = authn.get_jwt(request)
                username = authn.get_jwt_username(request, verify=False)
                email = authn.get_jwt_email(request, verify=False)

                logger.debug(f'User "{username}":"{email}" is currently admin; rerunning sync...')

                # Run their sync again to make absolutely sure they're still an admin
                user = django_auth.authenticate(request, token=token)
                if user and user.is_authenticated:
                    logger.debug('User has re-authenticated: {}'.format(username))

                    # Store this user in session
                    django_auth.login(request, user)

                    # Check updated status
                    if user.is_superuser or user.is_staff:
                        logger.debug(f'User "{username}":"{email}" is still admin')

                else:
                    logger.debug('User could not be authenticated: {}'.format(username))
                    # Whatever token this user has, it's not valid OR their account would/could not
                    # be created, deny permission. This will likely be the case for instances where
                    # automatic user creation is disabled and a user with a valid JWT is not being
                    # granted an account.
                    raise PermissionDenied

        return user


class DBMIJWTAuthenticationMiddleware(MiddlewareMixin):
    """
    This middleware does not use any user caching at all. For every request, the JWT, if present, is
    inspected to get the username and/or email and that is used to retrieve and set the user on
    the request object. Instead of caching the user's pk on first authentication and then using that
    to pull the user instance on subsequent requests, this simply authenticates the user on every request.
    Since the session is not used, instances of mismatched session and JWT should not occur. Instead
    of relying on the session to cache the current user, we simply use the JWT to do that.

    When to use this middleware: your site depends on JWT to authenticate and there's nothing heavy about the
    authorization process (syncing user state from other services, fetching authorizations, etc)
    """

    def process_request(self, request):
        request.user = SimpleLazyObject(lambda: self.__class__.get_jwt_user(request))

    @staticmethod
    def get_jwt_user(request):

        # Check for a valid token
        token = authn.get_jwt(request)
        if not token or not authn.validate_rs256_jwt(token):
            return AnonymousUser()

        # Get their username
        username = authn.get_jwt_username(request, verify=False)

        # Attempt to authenticate the current JWT.
        user = django_auth.authenticate(request, token=token)
        if not user or not user.is_authenticated:
            logger.debug('User could not be authenticated: {}'.format(username))
            # Whatever token this user has, it's not valid OR their account would/could not
            # be created, deny permission. This will likely be the case for instances where
            # automatic user creation is disabled and a user with a valid JWT is not being
            # granted an account.
            raise PermissionDenied

        return user



import os
import re
from io import open
from setuptools import find_packages, setup


def read(f):
    return open(f, 'r', encoding='utf-8').read()


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(os.path.dirname(os.path.abspath(__file__)), os.path.join(package, '__init__.py'))).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


version = get_version('dbmi_client')

setup(
    name='django-dbmi-client',
    version=version,
    url='https://github.com/hms-dbmi/django-dbmi-client',
    author='HMS DBMI Tech-core',
    author_email='dbmi-tech-core@hms.harvard.edu',
    packages=find_packages(exclude=['tests*']),
    license='Creative Commons Attribution-Noncommercial-Share Alike license',
    install_requires=read('requirements.txt').splitlines(),
    tests_require=read('requirements-test.txt').splitlines(),
    test_suite='dbmi_client.tests.runtests.main',
    include_package_data=True,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',  # example license
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



import mock
import re
import unittest
import responses

from django.http import HttpResponse, HttpRequest
from django.contrib.auth.models import User, AnonymousUser
from django.test.client import RequestFactory
from dbmi_client import authz
from dbmi_client.settings import dbmi_settings
from dbmi_client import auth


class TestAuth(unittest.TestCase):

    # User JWT
    fake_jwt = 'somefakejwtalsdijlaiwjdlijawlidjasdhgashgd'

    @classmethod
    def setUpClass(cls):

        # Patch JWT checker
        cls.jwt_patcher = mock.patch('dbmi_client.authn.validate_rs256_jwt')
        cls.mock_jwt = cls.jwt_patcher.start()
        cls.mock_jwt.return_value = True

        # Set a user and request factory
        cls.rf = RequestFactory()
        cls.user = User.objects.create_user(username='test_auth_user', email='testauthuser@dbmiauth.local', password='top_secret')

        # Set the authorization server URL
        cls.authz_url_pattern = re.compile(dbmi_settings.AUTHZ_URL + r'.*')

    @classmethod
    def tearDownClass(cls):

        # Disable patcher
        cls.jwt_patcher.stop()

    def build_request(self, path, method='get', user=True):

        # Build the request
        request = getattr(self.rf, method)(path, HTTP_AUTHORIZATION=TestAuth.fake_jwt)

        # Set empty session
        setattr(request, 'session', {})

        # Check for a user
        if user:
            # Add the user
            request.user = self.user
            request.COOKIES[dbmi_settings.JWT_COOKIE_NAME] = 'JWT {}'.format(TestAuth.fake_jwt)

        else:
            request.user = AnonymousUser()

        return request

    def setUp(self):

        # Setup each test
        pass

    @responses.activate
    @mock.patch('dbmi_client.authz.has_a_permission')
    def test_auth_mock_(self, mock_has_a_permission):

        # Set the response
        mock_has_a_permission.return_value = True

        # Build the request
        request = self.build_request('/some/page/')

        # Run it
        has_authz = authz.has_a_permission(request, self.user.email, 'item', ['admin', 'read'], check_parents=True)

        # Check it
        self.assertTrue(has_authz)

    @responses.activate
    def test_auth_response_(self):

        # Build the request
        request = self.build_request('/some/page/')

        # Build the response handler
        responses.add(responses.GET,
                      self.authz_url_pattern,
                      json={'error': 'not found'},
                      status=404)

        # Build the call
        has_authz = authz.has_permission(request, self.user.email, 'item', 'admin', check_parents=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertFalse(has_authz)

    @responses.activate
    def test_auth_dbmi_user_auth(self):

        # Build the request
        request = self.build_request('/some/page/', user=True)

        # Create a fake view
        view = mock.MagicMock(return_value=HttpResponse('Some page content'))

        # Run.
        decorated = auth.dbmi_user(view)
        response = decorated(request, *[], **{})

        # Check.
        # View was called.
        view.assert_called_once_with(request, *[], **{})

        # Ensure the user is allowed
        self.assertEqual(response.status_code, 200)

    @responses.activate
    @mock.patch('dbmi_client.authn.login_redirect')
    def test_auth_dbmi_user_no_auth(self, mock_login_redirect):

        # Build the request
        request = self.build_request('/some/page/', user=False)

        # Create a fake view
        view = mock.MagicMock(return_value=HttpResponse('Some page content'))

        # Run.
        decorated = auth.dbmi_user(view)
        response = decorated(request, *[], **{})

        # Check.
        # View was never reached.
        view.assert_not_called()

        # Ensure the user is redirected to login.
        mock_login_redirect.assert_called_once_with(request)
